import pytest
from pathlib import Path
from conf_schema import create_user_blogs, fill_user_blogs, drop_user_blogs
import utils

@pytest.fixture
def li_dbpath(tmpdir):
    db_path = str(Path(tmpdir.strpath) / "blogdb.db")
    yield db_path

@pytest.fixture
def li_db(rconn, queries):
    create_user_blogs(rconn, queries)
    fill_user_blogs(rconn, queries)
    yield rconn
    drop_user_blogs(rconn, queries)



from .ageneric import AsyncGenericAdapter
from .pyformat import _replacer
from ..utils import VAR_REF


class AsyncPyFormatAdapter(AsyncGenericAdapter):
    """Convert from named to pyformat parameter style."""

    def process_sql(self, query_name, op_type, sql):
        """From named to pyformat."""
        return VAR_REF.sub(_replacer, sql)



import datetime
import aiosql
import pytest
import pytest_asyncio
import run_tests as t
import utils as u
import copy

try:
    import psycopg as db
    from psycopg.rows import dict_row
except ModuleNotFoundError:
    pytest.skip("missing driver: psycopg", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
    pytest.mark.skipif(not u.has_pkg("pytest_asyncio"), reason="no pytest_asyncio"),
]

@pytest.fixture(scope="module")
def driver():
    return "apsycopg"

@pytest.fixture(scope="module")
def date():
    return datetime.date

# FIXME merge?

# raw connection without database data
@pytest_asyncio.fixture
async def rconn(pg_params):
    conn = await db.AsyncConnection.connect(**pg_params)
    yield conn
    await conn.close()

# asynchronous tuple connection with database data
@pytest_asyncio.fixture
async def aconn(rconn, pg_db):
    yield rconn

# asynchronous dict connection with database data
@pytest_asyncio.fixture
async def dconn(pg_params, pg_db):
    params = copy.copy(pg_params)
    params["row_factory"] = dict_row
    conn = await db.AsyncConnection.connect(**params)
    yield conn
    await conn.close()

from run_tests import (
    run_async_sanity as test_sanity,
    run_async_record_query as test_record_query,
    run_async_parameterized_record_query as test_parameterized_record_query,
    run_async_parameterized_query as test_parameterized_query,
    run_async_select_one as test_select_one,
    run_async_select_value as test_select_value,
    run_async_delete as test_delete,
    run_async_execute_script as test_execute_script,
    run_async_record_class_query as test_record_class_query,
    run_async_methods as test_methods,
    run_async_select_cursor_context_manager as test_select_cursor_context_manager,
    run_async_insert_returning as test_insert_returning,
    run_async_insert_many as test_insert_many,
)

def test_version():
    assert db.__version__.startswith("3.")



import aiosql
import pytest
import run_tests as t
import utils

try:
    import apsw as db
except ModuleNotFoundError:
    pytest.skip("missing driver: apsw", allow_module_level=True)

pytestmark = [
    pytest.mark.sqlite3,
]

@pytest.fixture(scope="module")
def driver():
    return "apsw"

@pytest.fixture(scope="module")
def date():
    return t.todate

# driver does not seem to return row counts on !
@pytest.fixture(scope="module")
def expect():
    return -1

class APSWConnection(db.Connection):
    """APSW Connection wrapper with autocommit off."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._begin()

    def _begin(self):
        self.cursor().execute("BEGIN").close()

    def commit(self):  # pragma: no cover
        self.cursor().execute("COMMIT").close()
        self._begin()

    def _rollback(self):
        self.cursor().execute("ROLLBACK").close()

    def rollback(self):  # pragma: no cover
        self._rollback()
        self._begin()

    def close(self):
        self._rollback()
        super().close()

@pytest.fixture
def rconn(li_dbpath):
    conn = APSWConnection(li_dbpath)
    yield conn
    conn.close()

@pytest.fixture
def conn(li_db):
    yield li_db

@pytest.fixture
def dconn(conn):
    conn.setrowtrace(utils.dict_factory)
    return conn

from run_tests import (
    run_sanity as test_sanity,
	run_something as test_something,
	run_cursor as test_cursor,
	run_record_query as test_record_query,
	run_parameterized_query as test_parameterized_query,
	run_parameterized_record_query as test_parameterized_record_query,
	run_record_class_query as test_record_class_query,
	run_select_cursor_context_manager as test_select_cursor_context_manager,
	run_select_one as test_select_one,
    # FIXME not supported?
	# run_insert_returning as test_insert_returning,
	run_delete as test_delete,
	run_insert_many as test_insert_many,
	run_select_value as test_select_value,
	run_date_time as test_date_time,
	run_object_attributes as test_object_attributes,
	run_execute_script as test_execute_script,
	run_modulo as test_modulo,
)



import datetime
import aiosql
import pytest
import utils as u

try:
    import pgdb as db  # PyGreSQL DB-API driver
except ModuleNotFoundError:
    pytest.skip("missing driver: pygresql", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
]

@pytest.fixture(scope="module")
def driver():
    return "pygresql"

@pytest.fixture(scope="module")
def date():
    return datetime.date

@pytest.fixture
def rconn(pg_params):
    params = dict(pg_params)
    params["database"] = params["dbname"]
    del params["dbname"]
    if "port" in params:
        params["host"] += ":" + params["port"]
        del params["port"]
    u.log.debug(f"params: {params}")
    with db.connect(**params) as conn:
        yield conn

@pytest.fixture
def conn(pg_db):
    yield pg_db

# FIXME dconn

from run_tests import (
    run_sanity as test_sanity,
	run_something as test_something,
	run_cursor as test_cursor,
	# run_record_query as test_record_query,
	# run_parameterized_record_query as test_parameterized_record_query,
	run_parameterized_query as test_parameterized_query,
	run_record_class_query as test_record_class_query,
	run_select_cursor_context_manager as test_select_cursor_context_manager,
	run_select_one as test_select_one,
	run_insert_returning as test_insert_returning,
	run_delete as test_delete,
	run_insert_many as test_insert_many,
	run_select_value as test_select_value,
	run_date_time as test_date_time,
	run_object_attributes as test_object_attributes,
	run_execute_script as test_execute_script,
	run_modulo as test_modulo,
)



import datetime
import aiosql
import pytest
import run_tests as t
import utils as u

try:
    import psycopg as db
    from psycopg.rows import dict_row
except ModuleNotFoundError:
    pytest.skip("missing driver: psycopg", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
]

@pytest.fixture(scope="module")
def driver():
    return "psycopg"

@pytest.fixture(scope="module")
def date():
    return datetime.date

@pytest.fixture
def rconn(pg_params):
    with db.connect(**pg_params) as conn:
        yield conn

@pytest.fixture
def conn(pg_db):
    yield pg_db

@pytest.fixture
def dconn(pg_params, pg_db):
    with db.connect(**pg_params, row_factory=dict_row) as conn:
        yield conn

from run_tests import (
    run_sanity as test_sanity,
    run_something as test_something,
    run_cursor as test_cursor,
    run_record_query as test_record_query,
    run_parameterized_record_query as test_parameterized_record_query,
    run_parameterized_query as test_parameterized_query,
    run_select_one as test_select_one,
    run_select_value as test_select_value,
    run_modulo as test_modulo,
    run_delete as test_delete,
    run_date_time as test_date_time,
    run_execute_script as test_execute_script,
    run_object_attributes as test_object_attributes,
    run_record_class_query as test_record_class_query,
    run_select_cursor_context_manager as test_select_cursor_context_manager,
    run_insert_returning as test_insert_returning,
    run_insert_many as test_insert_many,
)

def test_version():
    assert db.__version__.startswith("3.")

def test_select_value_dict(dconn, queries):
    t.run_select_value(dconn, queries)



import datetime
import aiosql
import pytest

try:
    import duckdb as db
except ModuleNotFoundError:
    pytest.skip("missing driver: duckdb", allow_module_level=True)

pytestmark = [pytest.mark.duckdb]

@pytest.fixture(scope="module")
def driver():
    return "duckdb"

@pytest.fixture(scope="module")
def date():
    return datetime.date

@pytest.fixture
def conn(duckdb_conn):
    return duckdb_conn

# FIXME dconn

from run_tests import (
    run_sanity as test_sanity,
	run_something as test_something,
	run_cursor as test_cursor,
    # KO
	# run_record_query as test_record_query,
	# run_parameterized_record_query as test_parameterized_record_query,
	# run_record_class_query as test_record_class_query,
	run_parameterized_query as test_parameterized_query,
	run_select_cursor_context_manager as test_select_cursor_context_manager,
	run_select_one as test_select_one,
	run_insert_returning as test_insert_returning,
	run_delete as test_delete,
	run_insert_many as test_insert_many,
	run_select_value as test_select_value,
	run_date_time as test_date_time,
	run_object_attributes as test_object_attributes,
	run_execute_script as test_execute_script,
	run_modulo as test_modulo,
)

def test_no_cursor(conn):
    adapter = aiosql.adapters.duckdb.DuckDBAdapter(use_cursor=False)
    cursor = adapter._cursor(conn)
    assert cursor == conn



import pytest
import importlib
import utils as u
import time
from conf_schema import create_user_blogs, fill_user_blogs, drop_user_blogs

def ms_has_db(conn, database):
    with conn.cursor() as cursor:
        cursor.execute("SELECT COUNT(*) AS cnt FROM sys.databases WHERE name = %s", (database,))
        return cursor.fetchone() in ({"cnt": 1}, (1,))

try:

    # NOTE rhetorical, there is only one ms driver for now
    @pytest.fixture(scope="module")
    def ms_driver(request):
        """Return driver class."""
        driver = request.config.getoption("mssql_driver") or "pymssql"
        db = importlib.import_module(driver)
        return db

    @pytest.fixture(scope="module")
    def ms_dsn(request):
        """Return connection parameters suitable to pymssql driver."""
        yield {
            # see conftest.py for list of options
            "server": request.config.getoption("mssql_server") or '.',
            "port": request.config.getoption("mssql_port") or 1433,
            "user": request.config.getoption("mssql_user") or "sa",
            "password": request.config.getoption("mssql_password"),
            "database": request.config.getoption("mssql_database") or "pytest",
            "as_dict": True,
            "autocommit": False,
        }

    @pytest.fixture(scope="module")
    def ms_master(ms_dsn):
        """Return connection parameters suitable for "system admin" access."""
        dsn = dict(ms_dsn)
        dsn["database"] = "master"
        dsn["autocommit"] = True
        yield dsn

    @pytest.fixture
    def ms_conn(request, ms_dsn, ms_driver):
        """Return a simple connection using the expected driver."""
        driver, db = ms_driver.__name__, ms_driver
        tries = request.config.getoption("mssql_tries") or 3
        with u.db_connect(ms_driver, tries, **ms_dsn) as conn:
            yield conn

    @pytest.fixture(scope="module")
    def ms_db(ms_driver, ms_dsn, ms_master, queries):
        """Build the test database and return a connection to that."""
        with ms_driver.connect(**ms_master) as conn:
            # initial contents if needed
            if not ms_has_db(conn, "pytest"):
                with conn.cursor() as cur:
                    cur.execute("CREATE DATABASE pytest")
                    cur.execute("USE pytest")
                    conn.commit()
                create_user_blogs(conn, queries)
                fill_user_blogs(conn, queries)
            else:
                u.log.warning("skipping pytest schema creation")
        # connection to pytest possibly database created above
        with ms_driver.connect(**ms_dsn) as conn:
            yield conn
        # cleanup
        with ms_driver.connect(**ms_master) as conn:
            with conn.cursor() as cur:
                # u.log.warning("cleaning up pytest schema")
                cur.execute("DROP DATABASE pytest")
            conn.commit()

except ModuleNotFoundError:
    # provide empty fixtures to please pytest "parsing"

    @pytest.fixture
    def ms_driver():
        raise Exception("unimplemented fixture")

    @pytest.fixture
    def ms_dsn():
        raise Exception("unimplemented fixture")

    @pytest.fixture
    def ms_master():
        raise Exception("unimplemented fixture")

    @pytest.fixture
    def ms_conn():
        raise Exception("unimplemented fixture")

    @pytest.fixture
    def ms_db():
        raise Exception("unimplemented fixture")



import datetime
import pytest
import aiosql
import run_tests as t
import utils as u

try:
    import asyncpg
    import pytest_asyncio
except ModuleNotFoundError as m:
    pytest.skip(f"missing module: {m}", allow_module_level=True)

pytestmark = [
    pytest.mark.postgres,
    # pytest.mark.asyncio,
    pytest.mark.skipif(not u.has_pkg("pytest_postgresql"), reason="no pytest_postgresql"),
    pytest.mark.skipif(not u.has_pkg("pytest_asyncio"), reason="no pytest_asyncio"),
]

@pytest.fixture(scope="module")
def driver():
    return "asyncpg"

@pytest.fixture(scope="module")
def date():
    return datetime.date

@pytest_asyncio.fixture
async def rconn(pg_dsn):
    conn = await asyncpg.connect(pg_dsn)
    yield conn
    await conn.close()

@pytest_asyncio.fixture
async def aconn(pg_db):
    yield pg_db

@pytest_asyncio.fixture
async def dconn(aconn):
    # FIXME dict row?
    yield aconn

from run_tests import (
    run_async_sanity as test_async_sanity,
    run_async_record_query as test_async_record_query,
    run_async_parameterized_query as test_async_parameterized_query,
    run_async_parameterized_record_query as test_async_parameterized_record_query,
    run_async_record_class_query as test_async_record_class_query,
    run_async_select_cursor_context_manager as test_async_select_cursor_context_manager,
    run_async_select_one as test_async_select_one,
    run_async_select_value as test_async_select_value,
    run_async_insert_returning as test_async_insert_returning,
    run_async_delete as test_async_delete,
    run_async_insert_many as test_async_insert_many,
    run_async_execute_script as test_async_execute_script,
)

# TODO other pools?
@pytest.mark.asyncio
async def test_with_pool(pg_dsn, queries, pg_db):
    async with asyncpg.create_pool(pg_dsn) as pool:
        async with pool.acquire() as conn:
            await t.run_async_insert_returning(conn, queries, datetime.date)

@pytest.mark.asyncio
async def test_async_methods(pg_dsn, queries, pg_db):
    async with asyncpg.create_pool(pg_dsn) as pool:
        await t.run_async_methods(pool, queries)

@pytest.mark.asyncio
async def test_no_publish(aconn, queries):
    # TODO move in run
    no_publish = queries.f("blogs.no_publish")
    res = await no_publish(aconn)
    assert res is None

def test_many_replacements(pg_dsn, queries):
    """If the replacement was longer than the variable, bad SQL was generated.

    The variable replacement code had a bug that caused it to miscalculate where in the
    original string to put the placeholders.  The SQL below would produce a query that
    ended with "$8, $9, $10$11:k);" because of this bug.

    This test would fail before the bug was fixed and passes afterward.

    This issue was reported in https://github.com/nackjicholson/aiosql/issues/90.
    """

    sql = """
      -- name: test<!
      INSERT INTO table VALUES (:a, :b, :c, :d, :e, :f, :g, :h, :i, :j, :k);
    """
    actual = aiosql.from_str(sql, "asyncpg").test.sql
    expected = "INSERT INTO table VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);"
    assert actual == expected

@pytest.mark.asyncio
async def test_variable_replacement(aconn, queries):
    # Addresses bug reported in the following github issue:
    # https://github.com/nackjicholson/aiosql/issues/51
    #
    # When the test fails the users_sql below will fail to increment the $2
    # variable for lastname.
    search = queries.f("users.search")
    users_sql = "select username from users where firstname = $1 and lastname = $2;"
    assert search.sql == users_sql

    users_res = [ row async for row in search(aconn, title="John", lastname="Doe") ]
    assert users_res == [("johndoe",)]

    square = queries.f("misc.square")
    square_sql = "select $1::int * $1::int as squared;"
    assert square.sql == square_sql

    square_res = await square(aconn, val=42)
    assert square_res == 1764

def test_maybe_order_params():
    a = aiosql.adapters.asyncpg.AsyncPGAdapter()
    try:
        a.maybe_order_params("foo", "wrong-type-parameter")
        pytest.fail("exception should be raised")  # pragma: no cover
    except ValueError as e:
        assert "dict or tuple" in str(e)



import re
import inspect
from pathlib import Path
from typing import Type, Sequence, Any

from .utils import SQLParseException, SQLLoadException, VAR_REF, VAR_REF_DOT, log
from .types import QueryDatum, QueryDataTree, SQLOperationType, DriverAdapterProtocol

# identifies name definition comments
_QUERY_DEF = re.compile(r"--\s*name\s*:\s*")

# identifies record class definition comments
_RECORD_DEF = re.compile(r"--\s*record_class\s*:\s*(\w+)\s*")

# extract a valid query name followed by an optional operation spec
# FIXME this accepts "1st" but seems to reject "é"
_NAME_OP = re.compile(
    # query name
    r"^(?P<name>\w+)"
    # optional list of parameters (foo, bla) or ()
    r"(|\((?P<params>(\s*|\s*\w+\s*(,\s*\w+\s*)*))\))"
    # operation, empty for simple select
    r"(?P<op>(|\^|\$|!|<!|\*!|#))$"
)

# forbid numbers as first character
_BAD_PREFIX = re.compile(r"^\d")

# get SQL comment contents
_SQL_COMMENT = re.compile(r"\s*--\s*(.*)$")

# map operation suffixes to their type
_OP_TYPES = {
    "<!": SQLOperationType.INSERT_RETURNING,
    "*!": SQLOperationType.INSERT_UPDATE_DELETE_MANY,
    "!": SQLOperationType.INSERT_UPDATE_DELETE,
    "#": SQLOperationType.SCRIPT,
    "^": SQLOperationType.SELECT_ONE,
    "$": SQLOperationType.SELECT_VALUE,
    "": SQLOperationType.SELECT,
}

# extracting comments requires some kind of scanner
_UNCOMMENT = re.compile(
    # single quote strings
    r"(?P<squote>\'(\'\'|[^\'])*\')|"
    # double quote strings
    r'(?P<dquote>"(""|[^"])+")|'
    # one-line comment
    r"(?P<oneline>--.*?$)|"
    # multiline comments, excluding SQL hints
    r"|(?P<multiline>/\*(?!\+[\s\S]*?\*/)[\s\S]*?\*/)",
    re.DOTALL | re.MULTILINE,
)


def _remove_ml_comments(code: str) -> str:
    """Remove /* ... */ comments from code"""
    # identify commented regions to be removed
    rm = []
    for m in _UNCOMMENT.finditer(code):
        ml = m.groupdict()["multiline"]
        if ml:
            rm.append(m.span())
    # keep whatever else
    ncode, current = "", 0
    for start, end in rm:
        ncode += code[current:start]
        current = end
    # get tail
    ncode += code[current:]
    return ncode


def _preprocess_object_attributes(attribute, sql):
    """Substitute o.a by o<attribute>a and keep track of variables."""

    attributes = {}

    def _replace(m):
        gd = m.groupdict()
        if gd["dquote"] is not None:
            return gd["dquote"]
        elif gd["squote"] is not None:
            return gd["squote"]
        else:
            var, att = gd["var_name"].split(".", 1)
            var_name = var + attribute + att
            if var not in attributes:
                attributes[var] = {}
            if att not in attributes[var]:
                attributes[var][att] = var_name
            return f"{gd['lead']}:{var_name}"

    sql = VAR_REF_DOT.sub(_replace, sql)

    return sql, attributes


class QueryLoader:
    """Load Queries.

    This class holds the various utilities to read SQL files and build
    QueryDatum, which will be transformed as functions in Queries.

    - :param driver_adapter: driver name or class.
    - :param record_classes: nothing of dict.
    - :param attribute: string to insert in place of ``.``.
    """

    def __init__(
        self,
        driver_adapter: DriverAdapterProtocol,
        record_classes: dict[str, Any]|None,
        attribute: str|None = None,
    ):
        self.driver_adapter = driver_adapter
        self.record_classes = record_classes if record_classes is not None else {}
        self.attribute = attribute

    def _make_query_datum(
        self,
        query: str,
        ns_parts: list[str],
        floc: tuple[Path|str, int],
    ) -> QueryDatum:
        """Build a query datum.

        - :param query: the spec and name (``query-name!\n-- comments\nSQL;\n``)
        - :param ns_parts: name space parts, i.e. subdirectories of loaded files
        - :param floc: file name and lineno the query was extracted from
        """
        lines = [line.rstrip() for line in query.strip().splitlines()]
        qname, qop, qsig = self._get_name_op(lines[0])
        if re.search(r"[^A-Za-z0-9_]", qname):
            log.warning(f"non ASCII character in query name: {qname}")
        if len(lines) <= 1:
            raise SQLParseException(f"empty query for: {qname} at {floc[0]}:{floc[1]}")
        record_class = self._get_record_class(lines[1])
        sql, doc = self._get_sql_doc(lines[2 if record_class else 1 :])
        if re.search("(?s)^[\t\n\r ;]*$", sql):
            raise SQLParseException(f"empty sql for: {qname} at {floc[0]}:{floc[1]}")
        signature = self._build_signature(sql, qname, qsig)
        query_fqn = ".".join(ns_parts + [qname])
        if self.attribute:  # :u.a -> :u__a, **after** signature generation
            sql, attributes = _preprocess_object_attributes(self.attribute, sql)
        else:  # pragma: no cover
            attributes = None
        sql = self.driver_adapter.process_sql(query_fqn, qop, sql)
        return QueryDatum(query_fqn, doc, qop, sql, record_class, signature, floc, attributes, qsig)

    def _get_name_op(self, text: str) -> tuple[str, SQLOperationType, list[str]|None]:
        """Extract name, parameters and operation from spec."""
        qname_spec = text.replace("-", "_")
        matched = _NAME_OP.match(qname_spec)
        if not matched or _BAD_PREFIX.match(qname_spec):
            raise SQLParseException(f'invalid query name and operation spec: "{qname_spec}"')
        nameop = matched.groupdict()
        params, rawparams = None, nameop["params"]
        if rawparams is not None:
            params = [p.strip() for p in rawparams.split(",")]
            if params == ['']:  # handle "( )"
                params = []
        operation = _OP_TYPES[nameop["op"]]
        if params and operation == "#":  # pragma: no cover  # FIXME it is covered?
            raise SQLParseException(f'cannot use named parameters in SQL script: "{qname_spec}"')
        return nameop["name"], operation, params

    def _get_record_class(self, text: str) -> Type|None:
        """Extract record class from spec."""
        rc_match = _RECORD_DEF.match(text)
        rc_name = rc_match.group(1) if rc_match else None
        # TODO: Probably will want this to be a class, marshal in, and marshal out
        return self.record_classes.get(rc_name) if isinstance(rc_name, str) else None

    def _get_sql_doc(self, lines: Sequence[str]) -> tuple[str, str]:
        """Separate SQL-comment documentation and SQL code."""
        doc, sql = "", ""
        for line in lines:
            doc_match = _SQL_COMMENT.match(line)
            if doc_match:
                doc += doc_match.group(1) + "\n"
            else:
                sql += line + "\n"

        return sql.strip(), doc.rstrip()

    def _build_signature(self, sql: str, qname: str, sig: list[str]|None) -> inspect.Signature:
        """Return signature object for generated dynamic function."""
        # FIXME what about the connection?!
        params = [inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)]
        names = set()
        for match in VAR_REF.finditer(sql):
            gd = match.groupdict()
            if gd["squote"] or gd["dquote"]:
                continue
            name = gd["var_name"]
            if name.isdigit() or name in names:
                continue
            if sig is not None:  # optional parameter declarations
                if name not in sig:
                    raise SQLParseException(f"undeclared parameter name in query {qname}: {name}")
            names.add(name)
            params.append(
                inspect.Parameter(
                    name=name,
                    kind=inspect.Parameter.KEYWORD_ONLY,
                )
            )
        if sig is not None and len(sig) != len(names):
            unused = sorted(n for n in sig if n not in names)
            raise SQLParseException(f"unused declared parameter in query {qname}: {unused}")
        return inspect.Signature(parameters=params)

    def load_query_data_from_sql(
        self, sql: str, ns_parts: list[str], fname: Path|str = "<unknown>"
    ) -> list[QueryDatum]:
        """Load queries from a string."""
        usql = _remove_ml_comments(sql)
        qdefs = _QUERY_DEF.split(usql)
        # FIXME lineno is from the uncommented file
        lineno = 1 + qdefs[0].count("\n")
        data = []
        # first item is anything before the first query definition, drop it!
        for qdef in qdefs[1:]:
            data.append(self._make_query_datum(qdef, ns_parts, (fname, lineno)))
            lineno += qdef.count("\n")
        return data

    def load_query_data_from_file(
        self, path: Path, ns_parts: list[str] = [], encoding=None
    ) -> list[QueryDatum]:
        """Load queries from a file."""
        return self.load_query_data_from_sql(path.read_text(encoding=encoding), ns_parts, path)

    def load_query_data_from_dir_path(
        self, dir_path, ext=(".sql",), encoding=None
    ) -> QueryDataTree:
        """Load queries from a directory."""
        if not dir_path.is_dir():
            raise ValueError(f"The path {dir_path} must be a directory")

        def _recurse_load_query_data_tree(path, ns_parts=[], ext=(".sql",), encoding=None):
            query_data_tree = {}
            for p in path.iterdir():
                if p.is_file():
                    if p.suffix not in ext:
                        continue
                    for query_datum in self.load_query_data_from_file(
                        p, ns_parts, encoding=encoding
                    ):
                        query_data_tree[query_datum.query_name] = query_datum
                elif p.is_dir():
                    query_data_tree[p.name] = _recurse_load_query_data_tree(
                        p, ns_parts + [p.name], ext=ext, encoding=encoding
                    )
                else:  # pragma: no cover
                    # This should be practically unreachable.
                    raise SQLLoadException(f"The path must be a directory or file, got {p}")
            return query_data_tree

        return _recurse_load_query_data_tree(dir_path, ext=ext, encoding=encoding)



from contextlib import asynccontextmanager
from ..types import AsyncDriverAdapterProtocol

# it is unclear how generic is this class
class AsyncGenericAdapter(AsyncDriverAdapterProtocol):

    is_aio_driver = True

    def process_sql(self, query_name, op_type, sql):
        return sql  # pragma: no cover

    # this is an asynchronous generator
    async def select(self, conn, query_name, sql, parameters, record_class=None):
        cur = await conn.execute(sql, parameters)
        try:
            if record_class is not None:
                column_names = [c[0] for c in cur.description]
                for row in await cur.fetchall():
                    yield record_class(**dict(zip(column_names, row)))
            else:
                # psycopg 3.3: async for row in cur.results():
                for row in await cur.fetchall():
                    yield row
        finally:
            await cur.close()

    async def select_one(self, conn, query_name, sql, parameters, record_class=None):
        cur = await conn.execute(sql, parameters)
        result = await cur.fetchone()
        if result is not None and record_class is not None:
            column_names = [c[0] for c in cur.description]
            result = record_class(**dict(zip(column_names, result)))
        await cur.close()
        return result

    async def select_value(self, conn, query_name, sql, parameters):
        cur = await conn.execute(sql, parameters)
        result = await cur.fetchone()
        res = result[0] if result else None
        await cur.close()
        return res

    @asynccontextmanager
    async def select_cursor(self, conn, query_name, sql, parameters):
        cur = await conn.execute(sql, parameters)
        yield cur
        await cur.close()

    async def insert_returning(self, conn, query_name, sql, parameters):
        cur = await conn.execute(sql, parameters)
        result = await cur.fetchone()
        # res = result[0] if result else None
        await cur.close()
        return result

    async def insert_update_delete(self, conn, query_name, sql, parameters):
        cur = await conn.execute(sql, parameters)
        rc = cur.rowcount if hasattr(cur, "rowcount") else None
        await cur.close()
        return rc

    async def insert_update_delete_many(self, conn, query_name, sql, parameters):
        cur = conn.cursor()
        res = await cur.executemany(sql, parameters)
        await cur.close()
        return res

    async def execute_script(self, conn, sql):
        cur = await conn.execute(sql)
        msg = cur.statusmessage if hasattr(cur, "statusmessage") else "DONE"
        await cur.close()
        return msg
